#include <string.h>
#include <gtk/gtk.h>
#include "utils.h"
#include "helpers.h"


/* Helper function to smooth over the difficult GtkTreeView api. 
 * Takes a table {col_name="type",...}  where type is string, boolean */
 // watch out!!  not ordered like you want
int tree_view_set_columns(lua_State* L) {
	dbg("doing setcolumns\n");
	
	const char *colnames[MAXTABLECOLS];
	GType coltypes[MAXTABLECOLS];
	int ncols = 0;
	lua_pushnil(L);
	while (lua_next(L, 1) != 0 && ncols < MAXTABLECOLS) {  // key, value on stack
		const char *coltypename;
		colnames[ncols] = lua_tostring(L, -2);
		coltypename = lua_tostring(L, -1);
		if (!strcmp(coltypename, "string")) {
			coltypes[ncols] = G_TYPE_STRING;
		} else if (!strcmp(coltypename, "boolean")) {
			coltypes[ncols] = G_TYPE_BOOLEAN;
		} else {
			coltypes[ncols] = G_TYPE_STRING;
		}
		ncols++;
		lua_pop(L,1);
	}
	// create the model and save its pointer in the widget table
	GtkTreeModel* treestore = (GtkTreeModel*)gtk_tree_store_newv(ncols, coltypes);
	lua_pushlightuserdata(L, treestore );
	lua_setfield(L, lua_upvalueindex(1), "treestore");
	
	// get the address of the treeview and set its model to the one just made
	lua_pushvalue(L, lua_upvalueindex(1));
	lua_getfield(L, -1, "address");
	GtkTreeView* treeview = lua_touserdata(L, -1);
	gtk_tree_view_set_model (treeview, treestore);
	
	// make and add the columns
	int i;
	for(i = 0; i<ncols; i++) {
		GtkTreeViewColumn *column;
		if (coltypes[i] == G_TYPE_STRING) {
			column = gtk_tree_view_column_new_with_attributes(
			              colnames[i],
			              gtk_cell_renderer_text_new(),
			              "text", i,
			              NULL);
		} else if (coltypes[i] == G_TYPE_BOOLEAN) {
			column = gtk_tree_view_column_new_with_attributes(
			              colnames[i],
			              gtk_cell_renderer_toggle_new(),
			              "active", i,
			              NULL);
		}
		gtk_tree_view_append_column(treeview, column);
	}
	
	return 0;
}
	
int tree_view_dump_data(lua_State *L) {
	lua_pushvalue(L, lua_upvalueindex(1)); // get the widget table
	lua_getfield(L, -1, "treestore"); // get the treestore ptr
	dbg("forund treestore ptr %p\n", lua_touserdata(L, -1));
	return 0;
}

/* Set GtkTreeView data in bulk from a table of row tables:
 * { {"col1data", "col2data", parent="2:1"},{col1data,...}}
 * The parent entry can be omitted. It is a GtkTreePath, either
 * as a string or (if top-level) an integer.
 */
int tree_view_set_data(lua_State *L) {
	lua_pushvalue(L, lua_upvalueindex(1)); // get the widget table
	lua_getfield(L, -1, "treestore"); // get the treestore ptr
	GtkTreeModel* tree_store = (GtkTreeModel*)lua_touserdata(L, -1);
	GtkTreeIter iter, parent_iter;
	
	int ncols = gtk_tree_model_get_n_columns(tree_store);
	
	///GValue gv; g_value_init(&gv, G_TYPE_STRING);
	
	const char *parent_path_str;
	int rowno, colno;
	const char *strval;  // may not be a string in general -- use union.
	lua_pushnil(L);
	while (lua_next(L, 1) != 0) {  // key, value on stack
		rowno = lua_tonumber(L, -2);
		// row data table is tos
		
		lua_getfield(L, -1, "parent");
		parent_path_str = lua_tostring(L, -1);
		lua_pop(L,1);
		
		dbg("rowno %d parent %s\n", rowno, parent_path_str);

		GtkTreePath *parent_path;
		if (parent_path_str
		    && (parent_path = gtk_tree_path_new_from_string (parent_path_str))
		    && gtk_tree_model_get_iter(tree_store, &parent_iter, parent_path)
		   ){
			gtk_tree_store_append((GtkTreeStore *)tree_store,
			                                             &iter, &parent_iter);
		} else {
			gtk_tree_store_append((GtkTreeStore *)tree_store, &iter, NULL);
		}

		for (colno=1; colno <= ncols; colno++) {
			// this init keeps compiler happy, but breaks the opacity of GValue
			GValue gv = {0,{{0},{0}}}; 
			g_value_init(&gv, G_TYPE_STRING);
			 
			lua_rawgeti(L, -1, colno);
			strval = lua_tostring(L, -1); // or whatever the column demands
			lua_pop(L,1);
			
			dbg("r%d c%d %s\n\n", rowno, colno, strval);
			
			g_value_set_string(&gv, strval);
			
			gtk_tree_store_set_value((GtkTreeStore *)tree_store, &iter,
			                                                  colno-1, &gv);
		}
		/// next deal with differnt column types.
		/// check missing column (null string)  no parent -> zero  are correct)
		
		lua_pop(L,1); // remove row data table;
	}
	return 0;
}

/* Wrap the fiddly setting of a new cursor */
/* Call with a widget and a cursor type (enum - e.g. GDK_WATCH=150) */
/* Call with -1 to "use the cursor of its parent" (?=reset it) */
int l_set_cursor(lua_State *L) {
	lua_getfield(L, 1, "address");
	GtkWidget *widget = lua_touserdata(L, -1);
	int cursor_type = luaL_checkinteger(L, 2);
	
	/* If I understood how ref and unref worked in this case,
	 * I'd know if there is a little leak here */

	GdkWindow *gdkwin = gtk_widget_get_root_window(widget);
	if (cursor_type == -1) {
		gdk_window_set_cursor(gdkwin, 0); // reset default, I hope
	} else {
		GdkCursor *cursor = gdk_cursor_new(cursor_type);
		gdk_window_set_cursor(gdkwin, cursor);
	}
	return 0;
}

void set_cursor_types(lua_State *L) {
	defconst(GDK_WATCH);
	defconst(GDK_LEFT_PTR);
	defconst(GDK_X_CURSOR);
	defconst(GDK_HAND2);
	defconst(GDK_PIRATE);
}

// Helper for text view
int l_text_view_insert_at_line(lua_State *L) {  // txt, lin_no
	GtkTextIter it;
	const gchar *text = luaL_checkstring(L, 1);
	gint line_number = luaL_checkinteger(L, 2);
	lua_pushvalue(L, lua_upvalueindex(1)); // get the addr of the buff
	GtkTextBuffer *buf = lua_touserdata(L, -1);
	gtk_text_buffer_get_iter_at_line(buf, &it, line_number);
	gtk_text_buffer_insert(buf, &it, text, -1);
	return 0;
}
// Add text to end and scroll into view
int l_text_view_add(lua_State *L) {  // txt
	const gchar *text = luaL_checkstring(L, 1);
	
	lua_pushvalue(L, lua_upvalueindex(1)); // get the addr of the txtview
	GtkTextView *tv = lua_touserdata(L, -1);
	
	lua_pushvalue(L, lua_upvalueindex(2)); // get the addr of the buff
	GtkTextBuffer *buf = lua_touserdata(L, -1);
	
	gtk_text_buffer_insert_at_cursor(buf, text, -1);
	
	GtkTextMark *mark = gtk_text_buffer_get_insert(buf);
	gtk_text_view_scroll_mark_onscreen (tv, mark);
	return 0;
}


/* Handle outstanding events from within a callback */
int l_do_events(lua_State * UNUSED(L)) {
	while (gtk_events_pending()) {
		gtk_main_iteration ();
	}
	return 0;
}

/* Check for bit set in bitmask */
int l_isset(lua_State *L) {
	unsigned int bit = luaL_checkinteger(L, 1);
	unsigned int mask = luaL_checkinteger(L, 2);
	lua_pushboolean(L, ((bit & mask) != 0) );
	return 1;
}


